library Blake2Hash;

{$R *.res}
{$E wdx}
{$IFDEF WIN64}
  {$E w64}
  {$IFNDEF VER230}
    {$E wdx64}
  {$ENDIF}
{$ENDIF}

{$WEAKLINKRTTI ON}
{$SETPEOPTFLAGS $160}

uses
  SysUtils,
  Windows,
  Classes,
  contplug,
  RTLConsts,

  { Hash functions from
    https://github.com/Xor-el/HashLib4Pascal }
  HlpHashFactory,
  HlpCRC, HlpHashSize,
  HlpIHash, HlpBlake3,
  HlpBlake2BParams, HlpIBlake2BParams,
  HlpBlake2SParams, HlpIBlake2SParams;

const
  cFieldsNames: array [0..4] of AnsiString = ('blake2bp',
    'blake2sp', 'blake2b', 'blake2s', 'blake3');

  cFieldsTypes: array [0..High(cFieldsNames)] of Integer =
    (ft_StringW, ft_StringW, ft_StringW, ft_StringW, ft_StringW);

  cFieldsUnits: AnsiString = '128|160|192|224|256|288|384|512';
  cFieldsUnitsBlake2s: AnsiString = '128|160|192|224|256';

var
  RTL: _RTL_CRITICAL_SECTION;


function FileOpenAsReadOnly(const sFileName: string): THandle;
begin
  Result := CreateFile(PChar(sFileName), GENERIC_READ, FILE_SHARE_READ or
    FILE_SHARE_WRITE or FILE_SHARE_DELETE, nil, OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL, 0);
  if Result = INVALID_HANDLE_VALUE then
    raise EFOpenError.CreateResFmt(@SFOpenErrorEx, [ExpandFileName(sFileName),
      SysErrorMessage(GetLastError)])
end;


function Calculate(const AHashInstance: IHash; const Stream: TStream): string;
begin
  Stream.Position := 0;
  Result := LowerCase(AHashInstance.ComputeStream(Stream).ToString);
end;

function ProcessFile(const FileName: string; FieldIndex, UnitIndex: Integer): string;
var
  FS: TFileStream;
  HashSize: THashSize;
begin
  Result := '';
  FS := TFileStream.Create(FileOpenAsReadOnly(FileName));
  try
    case UnitIndex of
      0: HashSize := THashSize.hsHashSize128;
      1: HashSize := THashSize.hsHashSize160;
      2: HashSize := THashSize.hsHashSize192;
      3: HashSize := THashSize.hsHashSize224;
      4: HashSize := THashSize.hsHashSize256;
      5: HashSize := THashSize.hsHashSize288;
      6: HashSize := THashSize.hsHashSize384;
      7: HashSize := THashSize.hsHashSize512;
    else
      HashSize := THashSize.hsHashSize256;
    end; // case

    case FieldIndex of
      0: Result := Calculate(THashFactory.TCrypto.CreateBlake2BP(Int32(HashSize), nil), FS);
      1: Result := Calculate(THashFactory.TCrypto.CreateBlake2SP(Int32(HashSize), nil), FS);
      2: Result := Calculate(THashFactory.TCrypto.CreateBlake2B
          (TBlake2BConfig.Create(HashSize) as IBlake2BConfig), FS);
      3: Result := Calculate(THashFactory.TCrypto.CreateBlake2S
          (TBlake2SConfig.Create(HashSize) as IBlake2SConfig, nil), FS);
      4: Result := Calculate(TBlake3.Create(HashSize, nil) as IHash, FS);
    end; // case

  finally
    FS.Free;
  end;
end;

function ContentGetSupportedField(FieldIndex: Integer; FieldName: PAnsiChar;
  Units: PAnsiChar; MaxLen: Integer): Integer; stdcall;
begin
  if FieldIndex in [0..High(cFieldsNames)] then
  begin
    StrCopy(Units,'');
    StrLCopy(FieldName, PAnsiChar(cFieldsNames[FieldIndex]), MaxLen - 1);
    Result := cFieldsTypes[FieldIndex];
    StrLCopy(Units, PAnsiChar(''), MaxLen - 1);
    if SameText(string(cFieldsNames[FieldIndex]), 'blake2s') or
      SameText(string(cFieldsNames[FieldIndex]), 'blake2sp')  then
      StrLCopy(Units, PAnsiChar(cFieldsUnitsBlake2s), MaxLen - 1)
    else
      StrLCopy(Units, PAnsiChar(cFieldsUnits), MaxLen - 1);
  end else
    Result := ftNoMoreFields;
end;

function GetValueW(FileName: PWideChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer;
var
  AFileName, AStr: string;
begin
  Result := ft_fieldempty;

  AFileName := FileName;
  if Length(AFileName) > MAX_PATH then
    if Pos('\\', AFileName) > 0 then
      AFileName := '\\?\UNC\' + Copy(AFileName, 3, Length(AFileName) - 2)
    else
      AFileName := '\\?\' + AFileName;

  if not FileExists(AFileName) then
    Exit;

  AStr := '';

  try
    AStr := ProcessFile(AFileName, FieldIndex, UnitIndex);
  except
  end;

  if AStr = '' then
    Exit;


  StrLCopy(PWideChar(FieldValue), PWideChar(AStr), MaxLen div 2 - 2);

  if FieldIndex in [0 .. High(cFieldsNames)] then
    Result := cFieldsTypes[FieldIndex]
  else
    Result := ft_fieldempty;
end;

function ContentGetValueW(FileName: PWideChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer; stdcall;
begin
  if (Flags and CONTENT_DELAYIFSLOW > 0) then
  begin
    Result := FT_DELAYED;
    Exit;
  end;

  EnterCriticalSection(RTL);
  try
    Result := GetValueW(FileName, FieldIndex, UnitIndex, FieldValue, MaxLen, Flags);
  finally
    LeaveCriticalSection(RTL);
  end;
end;

function ContentGetValue(FileName: PAnsiChar; FieldIndex, UnitIndex: Integer;
  FieldValue: PByte; MaxLen, Flags: Integer): Integer; stdcall;
begin
  Result := ContentGetValueW(PWideChar(WideString(AnsiString(FileName))),
    FieldIndex, UnitIndex, FieldValue, MaxLen, Flags)
end;

procedure DllInit(Code: Integer);
begin
  case Code of
    DLL_PROCESS_ATTACH:
    begin
    end;
    DLL_PROCESS_DETACH:
    begin
      DeleteCriticalSection(RTL);
    end;
  end; // case
end;

exports
  ContentGetSupportedField,
  ContentGetValue,
  ContentGetValueW;

begin
  //ReportMemoryLeaksOnShutdown := True;
  InitializeCriticalSection(RTL);
  DllProc := @DllInit;
  DllInit(DLL_PROCESS_ATTACH);
end.
